
static void print_gpf_info(int level, struct xen_cpu *cpu, struct regs *regs)
{
    uint8_t *code = (void*)regs->rip;
    
    printk(level, "%s: vcpu %d, index 0x%x%s%s%s, "
	   "rflags %" PRIxREG ", cs:rip %" PRIxREG ":%" PRIxREG " "
	   "-> 0x%02x, 0x%02x, 0x%02x, 0x%02x,  0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	   __FUNCTION__, cpu->id, (int)(regs->error >> 3),
	   (regs->error & 0x04) ? ", TI"  : "",
	   (regs->error & 0x02) ? ", IDT" : "",
	   (regs->error & 0x01) ? ", EXT" : "",
	   regs->rflags, regs->cs, regs->rip,
	   code[0], code[1], code[2], code[3],
	   code[4], code[5], code[6], code[7]);
}

static void print_page_fault_info(int level, struct xen_cpu *cpu, struct regs *regs, ureg_t cr2)
{
    printk(level, "%s:%s%s%s%s%s%s, rip %" PRIxREG ", cr2 %" PRIxREG ", vcpu %d\n",
	   __FUNCTION__,
#if longmode
	   is_kernel(cpu) ? " [kernel-mode]" : " [user-mode]",
#else
	   "",
#endif
	   regs->error & 0x01 ? " preset"  : " nopage",
	   regs->error & 0x02 ? " write"   : " read",
	   regs->error & 0x04 ? " user"    : " kernel",
	   regs->error & 0x08 ? " reserved-bit"  : "",
	   regs->error & 0x10 ? " instr-fetch"   : "",
	   regs->rip, cr2, cpu->id);
}

static void print_registers(int level, struct regs *regs)
{
#if longmode
# define DR "%016" PRIxREG
#else
# define DR "%08" PRIxREG
#endif

    printk(level, "printing registers\n");
    printk(level, "  rax " DR " rbx " DR " rcx " DR " rdx " DR "\n",
	   regs->rax, regs->rbx, regs->rcx, regs->rdx);
    printk(level, "  rsi " DR " rdi " DR " rsp " DR " rbp " DR "\n",
	   regs->rsi, regs->rdi, regs->rsp, regs->rbp);
#if longmode
    printk(level, "  r8  " DR " r9  " DR " r10 " DR " r11 " DR "\n",
	   regs->r8, regs->r9, regs->r10, regs->r11);
    printk(level, "  r12 " DR " r13 " DR " r14 " DR " r15 " DR "\n",
	   regs->r12, regs->r13, regs->r14, regs->r15);
#endif
    printk(level, "  cs %04" PRIxREG " ss %04" PRIxREG "\n",
	   regs->cs, regs->ss);

#undef DR
}

static int fixup_extable(struct regs *regs)
{
    uintptr_t *ptr;

    for (ptr = _estart; ptr < _estop; ptr += 2) {
	if (ptr[0] != regs->rip)
	    continue;
	printk(2, "fixup: %" PRIxPTR " -> %" PRIxPTR "\n", ptr[0], ptr[1]);
	regs->rip = ptr[1];
	vminfo.faults[XEN_FAULT_PAGE_FAULT_FIX_EXTAB]++;
	return 1;
    }
    return 0;
}

/* --------------------------------------------------------------------- */

static struct descriptor_32 mkdesc(const struct kvm_segment *seg)
{
    struct descriptor_32 desc;
    int shift = 0;

    shift  = seg->g ? 12 : 0;
    desc.a = (seg->limit >> shift) & 0xffff;
    desc.b = (seg->limit >> shift) & 0x000f0000;
    
    desc.a |= (seg->base & 0xffff) << 16;
    desc.b |= seg->base & 0xff000000;
    desc.b |= (seg->base & 0xff0000) >> 16;
    desc.b |= (seg->type & 0x0f) << 8;
    desc.b |= (seg->dpl & 0x03) << 13;

    if (seg->s)       desc.b |= (1 << 12);
    if (seg->present) desc.b |= (1 << 15);
    if (seg->avl)     desc.b |= (1 << 20);
    if (seg->l)       desc.b |= (1 << 21);
    if (seg->db)      desc.b |= (1 << 22);
    if (seg->g)       desc.b |= (1 << 23);

#if 0
    fprintf(stderr,
	    "%s: [%04x] %08llx +%08x type %x dpl %x [ %s %s %s %s %s %s] -> %08x:%08x\n",
	    __FUNCTION__, seg->selector,
	    seg->base, seg->limit, seg->type, seg->dpl,
	    seg->s       ? "s"  : " ",
	    seg->present ? "p"  : " ",
	    seg->avl     ? "a"  : " ",
	    seg->l       ? "l"  : " ",
	    seg->db      ? "db" : "  ",
	    seg->g       ? "g"  : " ",
	    desc.b, desc.a);
#endif

    return desc;
}

static inline void gdt_set(struct descriptor_32 *gdt, const struct kvm_segment *seg)
{
    gdt[ seg->selector >> 3 ] = mkdesc(seg);
}

static void cr_init(struct xen_cpu *cpu)
{
    ureg_t cr0, cr4;

    printk(2, "%s: cpu %d\n", __FUNCTION__, cpu->id);

    cr0  = read_cr0();
    cr0 |= X86_CR0_PE | X86_CR0_MP | X86_CR0_ET | X86_CR0_NE | \
	X86_CR0_WP | X86_CR0_AM | X86_CR0_PG;
    cr0 &= ~(X86_CR0_TS|X86_CR0_CD|X86_CR0_NW);
    print_bits("cr0", read_cr0(), cr0, cr0_bits);
    write_cr0(cr0);

    cr4 = read_cr4();
    print_bits("cr4", read_cr4(), cr4, cr4_bits);
    write_cr4(cr4);
}

static void stack_init(struct xen_cpu *cpu)
{
    uintptr_t *ptr;
    int pages;

    if (cpu->stack_low)
	return;

    /* allocate stack */
    pages = (boot_stack_high - boot_stack_low + PAGE_SIZE -1) / PAGE_SIZE;
    cpu->stack_low  = get_pages(pages, "stack");
    cpu->stack_high = cpu->stack_low + pages * PAGE_SIZE;

    /* set per-cpu data pointer */
    ptr = STACK_PTR(cpu, cpu_ptr);
    *ptr = (uintptr_t)cpu;

#if longmode
    /* copy and setup syscall trampoline from boot stack */
    memcpy(STACK_PTR(cpu, trampoline_start),
	   trampoline_start, trampoline_stop - trampoline_start);
    ptr = STACK_PTR(cpu, trampoline_patch);
    *ptr = (uintptr_t)trampoline_syscall;
#endif
}

void idt_load(void)
{
    struct {
	uint16_t  len;
	uintptr_t ptr;
    } __attribute__((packed)) idtp = {
	.len = sizeof(xen_idt)-1,
	.ptr = (uintptr_t)xen_idt,
    };

    asm volatile("lidt %0" : : "m" (idtp));
}

static struct xen_cpu *cpu_init(int id)
{
    struct xen_cpu *cpu;
    int pages = (sizeof(*cpu) + PAGE_SIZE -1) / PAGE_SIZE;
	
    cpu = get_pages(pages, "per-cpu data");
    cpu->id = id;
    cpu->vcpu_info = &shared_info.vcpu_info[id];
    cpu->vcpu_info->evtchn_upcall_mask = 1;

    gdt_init(cpu);
    stack_init(cpu);
    tss_init(cpu);
    idt_load();
    cr_init(cpu);
    msrs_init(cpu);
    pv_init(cpu);

    return cpu;
}

/* --------------------------------------------------------------------- */

int panic(char *message)
{
    printk(0, "panic: %s\n", message);
    vmexit_panic();
    return 0;
}

asmlinkage void do_illegal_instruction(struct regs regs)
{
    struct xen_cpu *cpu = get_cpu();
    int skip;

    vminfo.faults[XEN_FAULT_ILLEGAL_INSTRUCTION]++;
    if (context_is_emu(vcpu, &regs)) {
	print_registers(0, &regs);
	panic("ring0 (emu) illegal instruction");
    }
    if (context_is_user(cpu, &regs)) {
	uint8_t *i = (void*)regs.rip;
	printk(1, "user ill: at %p"
	       "  0x%02x, 0x%02x, 0x%02x, 0x%02x,"
	       "  0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	       i, i[0], i[1], i[2], i[3], i[4], i[5], i[6], i[7]);
	bounce_trap(cpu, &regs, 6, -1);
	return;
    }
    
    skip = emulate(cpu, &regs);
    switch (skip) {
    case -1: /* error */
	print_registers(0, &regs);
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	bounce_trap(cpu, &regs, 6, -1);
	break;
    default: /* handled */
	regs.rip += skip;
	break;
    }
}

asmlinkage void do_general_protection(struct regs regs)
{
    struct xen_cpu *cpu = get_cpu();
    int skip;

    vminfo.faults[XEN_FAULT_GENERAL_PROTECTION]++;
    if (context_is_emu(vcpu, &regs)) {
	if (fixup_extable(&regs))
	    return;
	print_gpf_info(0, cpu, &regs);
	print_registers(0, &regs);
	panic("ring0 (emu) general protection fault");
    }
    if (context_is_user(cpu, &regs)) {
	vminfo.faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	print_gpf_info(1, cpu, &regs);
	bounce_trap(cpu, &regs, 13, -1);
	return;
    }
    
    if (0 != regs.error) {
	print_gpf_info(0, cpu, &regs);
	print_registers(0, &regs);
	panic("unhandled kernel gpf");
    }

    skip = emulate(cpu, &regs);
    switch (skip) {
    case -1: /* error */
	print_registers(0, &regs);
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	vminfo.faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	bounce_trap(cpu, &regs, 13, -1);
	break;
    default: /* handled */
	vminfo.faults[XEN_FAULT_GENERAL_PROTECTION_EMUINS]++;
	regs.rip += skip;
	break;
    }
}

asmlinkage void do_guest_forward(struct regs regs)
{
    struct xen_cpu *cpu = get_cpu();
    const struct trapinfo *trap = NULL;

    if (regs.trapno < sizeof(trapinfo)/sizeof(trapinfo[0]))
	trap = trapinfo + regs.trapno;
    printk(trap ? trap->lvl : 0,
	   "%s: trap %d [%s], error 0x%" PRIxREG ","
	   " cs:rip %" PRIxREG ":%" PRIxREG ","
	   " forwarding to guest\n",
	   __FUNCTION__, (int)regs.trapno,
	   trap && trap->name ? trap->name : "-",
	   trap && trap->ec   ? regs.error : 0,
	   regs.cs, regs.rip);
    bounce_trap(cpu, &regs, regs.trapno, -1);
}

asmlinkage void do_lazy_fpu(struct regs regs)
{
    struct xen_cpu *cpu = get_cpu();

    vminfo.faults[XEN_FAULT_LAZY_FPU]++;
    clts();
    bounce_trap(cpu, &regs, regs.trapno, -1);
}
